@@ -8,6 +8,7 @@
 #include "prime_iterator.h"
 #include "ecpp.h"
 #include "factor.h"
+#include "isaac.h"
 
 #define FUNC_gcd_ui 1
 #define FUNC_mpz_logn 1
@@ -33,9 +34,9 @@ static const unsigned short sprimes[NSMALLPRIMES] = {2,3,5,7,11,13,17,19,23,29,3
 
 void _GMP_init(void)
 {
-  /* We should  not use this random number system for crypto, so
-   * using this lousy seed is ok.  We just would like something a
-   * bit different every run.  Using Perl_seed(aTHX) would be better. */
+  /* For real use of randomness we need to be seeded properly.
+   * This gives us a start until someone calls seed_csprng().
+   * We could try to improve this duct-tape in various ways. */
   unsigned long seed = time(NULL);
   init_randstate(seed);
   prime_iterator_global_startup();
@@ -486,51 +487,363 @@ void _GMP_primorial(mpz_t prim, UV n)
 #endif
 }
 
-/* Luschny's version of the "Brent-Harvey" method */
-void bernfrac(mpz_t num, mpz_t den, mpz_t zn)
-{
-  UV k, j, n = mpz_get_ui(zn);
-  mpz_t* T;
-  mpz_t t;
+/*****************************************************************************/
 
-  if      (n == 0) { mpz_set_ui(num, 1);  mpz_set_ui(den, 1); return; }
-  else if (n == 1) { mpz_set_ui(num, 1);  mpz_set_ui(den, 2); return; }
-  else if (n & 1)  { mpz_set_ui(num, 0);  mpz_set_ui(den, 1); return; }
-  n >>= 1;
+/* Put result into char with correct number of digits */
+static char* _str_real(mpf_t f, unsigned long prec) {
+  char* out;
+  unsigned long k;
+  int neg = (mpf_sgn(f) < 0);
 
-  /* Algorithm TangentNumbers from https://arxiv.org/pdf/1108.0286.pdf */
-  New(0, T, n+1, mpz_t);
-  for (k = 1; k <= n; k++)  mpz_init(T[k]);
-  mpz_set_ui(T[1], 1);
+  if (neg)
+    mpf_neg(f, f);
 
-  mpz_init(t);
+  for (k = 0;  mpf_cmp_ui(f, 1000000000U) >= 0;  k += 9)
+    mpf_div_ui(f, f, 1000000000U);
+  for (;  mpf_cmp_ui(f, 1) >= 0;  k++)
+    mpf_div_ui(f, f, 10);
 
-  for (k = 2; k <= n; k++)
-    mpz_mul_ui(T[k], T[k-1], k-1);
+  New(0, out, 10+((k>prec) ? k : prec), char);
+  gmp_sprintf(out, "%.*Ff", prec, f);
+  memmove(out, out+2, prec);
 
-  for (k = 2; k <= n; k++) {
-    for (j = k; j <= n; j++) {
-      mpz_mul_ui(t, T[j], j-k+2);
-      mpz_mul_ui(T[j], T[j-1], j-k);
-      mpz_add(T[j], T[j], t);
+  if (k >= prec) { /* No decimal */
+    if (k-prec < 10) {
+      memset(out+prec, '0', k-prec);
+      prec=k-1;
+    } else {
+      out[prec++] = 'E';
+      prec += sprintf(out+prec, "%lu", k-prec+1);
     }
+  } else {        /* insert decimal in correct place */
+    memmove(out+k+1, out+k, prec-k);
+    out[k] = '.';
+  }
+  out[prec+1]='\0';
+  if (neg) {
+    memmove(out+1, out, prec+2);
+    out[0] = '-';
+  }
+//printf("out=%s\n",out);
+  return out;
+}
+
+static char* _frac_real(mpz_t num, mpz_t den, unsigned long prec) {
+#if 0
+  char* out;
+  mpf_t fnum, fden, res;
+  unsigned long numbits = mpz_sizeinbase(num,  2);
+  unsigned long denbits = mpz_sizeinbase(den,  2);
+  unsigned long numdigs = mpz_sizeinbase(num, 10);
+  unsigned long dendigs = mpz_sizeinbase(den, 10);
+
+  mpf_init2(fnum, 1 + numbits);  mpf_set_z(fnum, num);
+  mpf_init2(fden, 1 + denbits);  mpf_set_z(fden, den);
+  mpf_init2(res, (unsigned long) (8 + (numbits-denbits+1) + prec*3.4) );
+  mpf_div(res, fnum, fden);
+  mpf_clear(fnum);  mpf_clear(fden);
+
+  New(0, out, (10+numdigs-dendigs)+prec, char);
+  gmp_sprintf(out, "%.*Ff", (int)(prec), res);
+  mpf_clear(res);
+
+  return out;
+#else
+  char* out;
+  mpf_t fnum, fden;
+  unsigned long bits = 32+(unsigned long)(prec*3.32193);
+  mpf_init2(fnum, bits);   mpf_set_z(fnum, num);
+  mpf_init2(fden, bits);   mpf_set_z(fden, den);
+  mpf_div(fnum, fnum, fden);
+  out = _str_real(fnum, prec);
+  mpf_clear(fden);
+  mpf_clear(fnum);
+  return out;
+#endif
+}
+
+
+/*********************     Riemann Zeta and Riemann R     *********************/
+
+static void _bern_real_zeta(mpf_t bn, mpz_t zn, unsigned long prec);
+static unsigned long zeta_n = 0;
+static mpz_t* zeta_d = 0;
+
+static void _borwein_d(unsigned long D) {
+  mpz_t t1, t2, t3, sum;
+  unsigned long i, n = 3 + (1.31 * D);
+
+  if (zeta_n >= n)
+    return;
+
+  if (zeta_n > 0) {
+    for (i = 0; i <= zeta_n; i++)
+      mpz_clear(zeta_d[i]);
+    Safefree(zeta_d);
   }
 
-  /* (14), also last line of Algorithm FastTangentNumbers from paper */
-  mpz_mul_ui(num, T[n], n);
-  mpz_mul_si(num, num, (n & 1) ? 2 : -2);
-  mpz_set_ui(t, 1);
-  mpz_mul_2exp(den, t, 2*n);  /* den = U = 1 << 2n  */
-  mpz_sub_ui(t, den, 1);      /* t = U-1            */
-  mpz_mul(den, den, t);       /* den = U*(U-1)      */
-  mpz_gcd(t, num, den);
-  mpz_divexact(num, num, t);
-  mpz_divexact(den, den, t);
+  n += 10;   /* Add some in case we want a few more digits later */
+  zeta_n = n;
+  New(0, zeta_d, n+1, mpz_t);
+  mpz_init(t1); mpz_init(t2); mpz_init(t3);
 
-  mpz_clear(t);
-  for (k = 1; k <= n; k++)  mpz_clear(T[k]);
-  Safefree(T);
+  mpz_init_set_ui(sum, 1);
+  mpz_init_set(zeta_d[0], sum);
+
+  mpz_fac_ui(t1, n);
+  mpz_fac_ui(t2, n);
+  for (i = 1; i <= n; i++) {
+    mpz_mul_ui(t1, t1, 2*(n+i-1));    /* We've pulled out a 2 from t1 and t2 */
+    mpz_divexact_ui(t2, t2, n-i+1);
+    mpz_mul_ui(t2, t2, (2*i-1) * i);
+    mpz_divexact(t3, t1, t2);
+    mpz_add(sum, sum, t3);
+    mpz_init_set(zeta_d[i], sum);
+  }
+  mpz_clear(sum); mpz_clear(t3); mpz_clear(t2); mpz_clear(t1);
 }
+
+/* MPFR does some shortcuts, then does an in-place version of Borwein 1991.
+ * It's quite clever, and has the advantage of not using the statics.  Our
+ * code can be a little faster in some cases, slower in others.  They
+ * certainly have done more rigorous error bounding, allowing fewer guard
+ * bits and earlier loop exits.
+ *
+ * For real values, we use our home-grown mpf_pow function, which is very
+ * slow at high precisions and again very crude compared to MPFR.
+ * For non-integer values this won't compare at all in speed to MPFR or Pari.
+ */
+
+static void _zeta(mpf_t z, mpf_t f, unsigned long prec)
+{
+  unsigned long k, S, p;
+  mpf_t s, tf, term;
+  mpz_t t1;
+
+  if (mpf_cmp_ui(f,1) == 0) {
+   mpf_set_ui(z, 0);
+   return;
+ }
+
+  /* Shortcut if we know all prec terms are zeros. */
+  if (mpf_cmp_ui(f, 1+3.3219281*prec) >= 0 || mpf_cmp_ui(f, mpf_get_prec(z)) > 0) {
+    mpf_set_ui(z,1);
+    return;
+  }
+
+  S = (mpf_integer_p(f) && mpf_fits_ulong_p(f))  ?  mpf_get_ui(f)  :  0;
+
+  /* Negative integers using Bernoulli */
+  if (S == 0 && mpf_integer_p(f) && mpf_fits_slong_p(f) && mpf_sgn(f) != 0) {
+    S = -mpf_get_si(f);
+    if (!(S & 1)) { /* negative even integers are zero */
+      mpf_set_ui(z,0);
+    } else {        /* negative odd integers are -B_(n+1)/(n+1) */
+      mpz_t n;
+      mpz_init_set_ui(n, S+1);
+      _bern_real_zeta(z, n, prec);
+      mpf_div_ui(z, z, S+1);
+      mpf_neg(z,z);
+      mpz_clear(n);
+    }
+    return;
+  }
+
+  mpf_init2(s,    mpf_get_prec(z));   mpf_set(s, f);
+  mpf_init2(tf,   mpf_get_prec(z));
+  mpf_init2(term, mpf_get_prec(z));
+  mpz_init(t1);
+
+  if (S && S <= 14 && !(S & 1)) {         /* Small even S can be done with Pi */
+    unsigned long div[]={0,6,90,945,9450,93555,638512875,18243225};
+    char* pi = pidigits(prec+8);
+    mpf_set_str(z, pi, 10);
+    mpf_pow_ui(z, z, S);
+    if (S == 12) mpf_mul_ui(z, z, 691);
+    if (S == 14) mpf_mul_ui(z, z, 2);
+    mpf_div_ui(z, z, div[S/2]);
+    Safefree(pi);
+  } else if (mpf_cmp_ui(f, 3+prec*2.15) > 0) {  /* Only one term (3^s < prec) */
+    if (S) {
+      mpf_set_ui(term, 1);
+      mpf_mul_2exp(term, term, S);
+    } else {
+      mpf_set_ui(term, 2);
+      mpf_pow(term, term, s);
+    }
+    mpf_sub_ui(tf, term, 1);
+    mpf_div(z, term, tf);
+  } else if ( (mpf_cmp_ui(f,20) > 0 && mpf_cmp_ui(f, prec/3.5) > 0) ||
+              (prec > 500 && (mpz_ui_pow_ui(t1, 8*prec, S), mpz_sizeinbase(t1,2) > (20+3.3219281*prec))) ) {
+    /* Basic formula, for speed (also note only valid for > 1) */
+    PRIME_ITERATOR(iter);
+    mpf_set_ui(z, 1);
+    for (p = 2; p <= 1000000000; p = prime_iterator_next(&iter)) {
+      if (S) {
+        mpz_ui_pow_ui(t1, p, S);
+        mpf_set_z(term, t1);
+      } else {
+        mpf_set_ui(tf, p);
+        mpf_pow(term, tf, s);
+        mpz_set_f(t1, term);
+      }
+      if (mpz_sizeinbase(t1,2) > (20+3.3219281*prec)) break;
+      mpf_sub_ui(tf, term, 1);
+      mpf_div(term, term, tf);
+      mpf_mul(z, z, term);
+    }
+    prime_iterator_destroy(&iter);
+  } else {
+    /* TODO: negative non-integer inputs past -20 or so are very wrong. */
+    _borwein_d( (mpf_cmp_d(f,-3.0) >= 0)  ?  prec  :  80+2*prec );
+
+    mpf_set_ui(z, 0);
+    for (k = 0; k <= zeta_n-1; k++) {
+      if (S) {
+        mpz_ui_pow_ui(t1, k+1, S);
+        mpf_set_z(term, t1);
+      } else {
+        mpf_set_ui(tf, k+1);
+        mpf_pow(term, tf, s);
+      }
+
+      mpz_sub(t1, zeta_d[k], zeta_d[zeta_n]);
+      mpf_set_z(tf, t1);
+
+      mpf_div(term, tf, term);
+
+      if (k&1) mpf_sub(z, z, term);
+      else     mpf_add(z, z, term);
+    }
+
+    mpf_set_z(tf, zeta_d[zeta_n]);
+    mpf_div(z, z, tf);
+
+    if (S) {
+      mpf_set_ui(tf, 1);
+      mpf_div_2exp(tf, tf, S-1);
+    } else {
+      mpf_set_ui(term, 2);
+      mpf_ui_sub(tf, 1, s);
+      mpf_pow(tf, term, tf);
+    }
+
+    mpf_ui_sub(tf, 1, tf);
+    mpf_div(z, z, tf);
+
+    mpf_neg(z, z);
+  }
+  mpz_clear(t1);
+  mpf_clear(term); mpf_clear(tf); mpf_clear(s);
+}
+
+static void _zetaint(mpf_t z, unsigned long s, unsigned long prec)
+{
+  mpf_t f;
+
+  if (s <= 1) {
+    mpf_set_ui(z, 0);
+  } else if (s >= (1+3.3219281*prec) || s > mpf_get_prec(z)) {
+    /* Shortcut if we know all prec terms are zeros. */
+    mpf_set_ui(z,1);
+  } else {
+    mpf_init2(f, mpf_get_prec(z));
+    mpf_set_ui(f, s);
+    _zeta(z, f, prec);
+    mpf_clear(f);
+  }
+}
+
+static void _riemann_r(mpf_t r, mpf_t n, unsigned long prec)
+{
+  mpf_t logn, sum, term, part_term, tol, tf;
+  unsigned long k, bits = mpf_get_prec(n);
+
+  mpf_init2(logn,      bits);
+  mpf_init2(sum,       bits);
+  mpf_init2(term,      bits);
+  mpf_init2(part_term, bits);
+  mpf_init2(tol,       bits);
+  mpf_init2(tf,        bits);
+
+  mpf_log(logn, n);
+  mpf_set_ui(tol, 10);  mpf_pow_ui(tol, tol, prec);  mpf_ui_div(tol,1,tol);
+
+#if 1 /* Standard Gram Series */
+  mpf_set_ui(part_term, 1);
+  mpf_set_ui(sum, 1);
+  for (k = 1; k < 10000; k++) {
+    mpf_mul(part_term, part_term, logn);
+    mpf_div_ui(part_term, part_term, k);
+
+    _zetaint(tf, k+1, prec+1);
+    mpf_mul_ui(tf, tf, k);
+    mpf_div(term, part_term, tf);
+    mpf_add(sum, sum, term);
+
+    mpf_abs(term, term);
+    mpf_mul(tf, sum, tol);
+    if (mpf_cmp(term, tf) <= 0) break;
+  }
+#else  /* Accelerated (about half the number of terms needed) */
+  /* See:   http://mathworld.wolfram.com/GramSeries.html (5)
+   * Ramanujan's G (3) and its restatements (5) and (6) are equal,
+   * but G is only asymptotically equal to R (restated in (4) as per Gram).
+   * To avoid confusion we won't use this.  Too bad, as it can be 2x faster.
+   */
+  mpf_set(part_term, logn);
+  mpf_set_ui(sum, 0);
+  _zetaint(tf, 2, prec);
+  mpf_div(term, part_term, tf);
+  mpf_add(sum, sum, term);
+  for (k = 2; k < 1000000; k++) {
+    if (mpf_cmp(part_term, tol) <= 0) break;
+
+    mpf_mul(tf, logn, logn);
+    if (k < 32768) {
+      mpf_div_ui(tf, tf, (2*k-2) * (2*k-1));
+    } else {
+      mpf_div_ui(tf, tf, 2*k-2);  mpf_div_ui(tf, tf, 2*k-1);
+    }
+    mpf_mul(part_term, part_term, tf);
+
+    _zetaint(tf, 2*k, prec);
+    mpf_mul_ui(tf, tf, 2*k-1);
+    mpf_div(term, part_term, tf);
+    mpf_add(sum, sum, term);
+  }
+  mpf_mul_ui(sum, sum, 2);
+  mpf_add_ui(sum, sum, 1);
+#endif
+
+  mpf_set(r, sum);
+
+  mpf_clear(tf); mpf_clear(tol); mpf_clear(part_term);
+  mpf_clear(term); mpf_clear(sum); mpf_clear(logn);
+}
+
+
+char* zetareal(mpf_t z, unsigned long prec)
+{
+  size_t est_digits = 10+prec;
+  char* out;
+  if (mpf_cmp_ui(z,1) == 0) return 0;
+  if (mpz_sgn(z) < 0) est_digits += -mpf_get_si(z);
+  _zeta(z, z, prec);
+  New(0, out, est_digits, char);
+  gmp_sprintf(out, "%.*Ff", (int)(prec), z);
+  return out;
+}
+
+char* riemannrreal(mpf_t r, unsigned long prec)
+{
+  if (mpf_cmp_ui(r,0) <= 0) return 0;
+  _riemann_r(r, r, prec);
+  return _str_real(r, prec);
+}
+
+/***************************        Harmonic        ***************************/
+
 static void _harmonic(mpz_t a, mpz_t b, mpz_t t) {
   mpz_sub(t, b, a);
   if (mpz_cmp_ui(t, 1) == 0) {
@@ -564,51 +877,351 @@ void harmfrac(mpz_t num, mpz_t den, mpz_t zn)
   mpz_clear(t);
 }
 
-static char* frac_real(mpz_t num, mpz_t den, unsigned long prec) {
-  char* out;
-  mpf_t fnum, fden, res;
-  unsigned long numbits = mpz_sizeinbase(num,  2);
-  unsigned long denbits = mpz_sizeinbase(den,  2);
-  unsigned long numdigs = mpz_sizeinbase(num, 10);
-  unsigned long dendigs = mpz_sizeinbase(den, 10);
-
-  mpf_init2(fnum, 1 + numbits);  mpf_set_z(fnum, num);
-  mpf_init2(fden, 1 + denbits);  mpf_set_z(fden, den);
-  mpf_init2(res, (unsigned long) (8 + (numbits-denbits+1) + prec*3.4) );
-  mpf_div(res, fnum, fden);
-  mpf_clear(fnum);  mpf_clear(fden);
-
-  New(0, out, (10+numdigs-dendigs)+prec, char);
-  gmp_sprintf(out, "%.*Ff", (int)(prec), res);
-  mpf_clear(res);
-
-  return out;
-}
-
 char* harmreal(mpz_t zn, unsigned long prec) {
   char* out;
   mpz_t num, den;
 
   mpz_init(num); mpz_init(den);
   harmfrac(num, den, zn);
-  out = frac_real(num, den, prec);
+  out = _frac_real(num, den, prec);
   mpz_clear(den); mpz_clear(num);
 
   return out;
 }
 
+/**************************        Bernoulli        **************************/
+
+static void _bern_real_zeta(mpf_t bn, mpz_t zn, unsigned long prec)
+{
+  unsigned long s = mpz_get_ui(zn);
+  mpf_t tf;
+
+  if (s & 1) {
+    mpf_set_d(bn, (s == 1) ? 0.5 : 0.0);
+    return;
+  }
+
+  mpf_init2(tf, mpf_get_prec(bn));
+
+  /* For large values with low precision, we should look at approximations.
+   *   http://www.ebyte.it/library/downloads/2008_MTH_Nemes_GammaApproximationUpdate.pdf
+   *   http://www.luschny.de/math/primes/bernincl.html
+   *   http://arxiv.org/pdf/math/0702300.pdf
+   */
+
+  _zetaint(bn, s, prec);
+
+  /* We should be using an approximation here, e.g. Pari's mpfactr.  For
+   * large values this is the majority of time taken for this function. */
+  { mpz_t t; mpz_init(t); mpz_fac_ui(t, s); mpf_set_z(tf, t); mpz_clear(t);}
+  mpf_mul(bn, bn, tf);
+  /* bn = s! * zeta(s) */
+
+  { char* pi = pidigits(prec+8); mpf_set_str(tf, pi, 10); Safefree(pi); }
+  mpf_mul_ui(tf, tf, 2);
+  mpf_pow_ui(tf, tf, s);
+  mpf_div(bn, bn, tf);
+  /* bn = s! * zeta(s) / (2Pi)^s */
+
+  mpf_mul_2exp(bn, bn, 1);
+  if ((s & 3) == 0) mpf_neg(bn, bn);
+  /* bn = (-1)^(n-1) * 2 * s! * zeta(s) / (2Pi)^s */
+  mpf_clear(tf);
+}
+
+
+static void _bernfrac_comb(mpz_t num, mpz_t den, mpz_t zn, mpz_t t)
+{
+  unsigned long k, j, n = mpz_get_ui(zn);
+  mpz_t* T;
+
+  if (n <= 1 || (n & 1)) {
+    mpz_set_ui(num, (n<=1) ? 1 : 0);
+    mpz_set_ui(den, (n==1) ? 2 : 1);
+    return;
+  }
+
+  /* Denominator */
+  mpz_set_ui(t, 1);
+  mpz_mul_2exp(den, t, n);    /* den = U = 1 << n  */
+  mpz_sub_ui(t, den, 1);      /* t = U-1            */
+  mpz_mul(den, den, t);       /* den = U*(U-1)      */
+
+  n >>= 1;
+
+  /* Luschny's version of the "Brent-Harvey" method */
+  /* Algorithm TangentNumbers from https://arxiv.org/pdf/1108.0286.pdf */
+  New(0, T, n+1, mpz_t);
+  for (k = 1; k <= n; k++)  mpz_init(T[k]);
+  mpz_set_ui(T[1], 1);
+
+  for (k = 2; k <= n; k++)
+    mpz_mul_ui(T[k], T[k-1], k-1);
+
+  for (k = 2; k <= n; k++) {
+    for (j = k; j <= n; j++) {
+      mpz_mul_ui(t, T[j], j-k+2);
+      mpz_mul_ui(T[j], T[j-1], j-k);
+      mpz_add(T[j], T[j], t);
+    }
+  }
+
+  /* (14), also last line of Algorithm FastTangentNumbers from paper */
+  mpz_mul_ui(num, T[n], n);
+  mpz_mul_si(num, num, (n & 1) ? 2 : -2);
+
+  for (k = 1; k <= n; k++)  mpz_clear(T[k]);
+  Safefree(T);
+}
+
+
+static void _bernfrac_zeta(mpz_t num, mpz_t den, mpz_t zn, mpz_t t)
+{
+  unsigned long prec, n = mpz_get_ui(zn);
+  double nbits;
+  mpf_t bn, tf;
+  /* Compute integer numerator by getting the real bn first. */
+
+  if (n <= 1 || (n & 1)) {
+    mpz_set_ui(num, (n<=1) ? 1 : 0);
+    mpz_set_ui(den, (n==1) ? 2 : 1);
+    return;
+  }
+  if (n == 2) { mpz_set_ui(num, 1); mpz_set_ui(den, 6); return; }
+
+  /* Calculate denominator */
+  {
+    int i, ndivisors;
+    mpz_t *D;
+
+    mpz_set_ui(t, n >> 1);
+    D = divisor_list(&ndivisors, t);
+    mpz_set_ui(den, 6);
+    for (i = 1; i < ndivisors; i++) {
+      mpz_mul_2exp(t,D[i],1);  mpz_add_ui(t,t,1);
+      if (_GMP_is_prime(t))
+        mpz_mul(den, den, t);
+    }
+    for (i = 0; i < ndivisors; i++)
+      mpz_clear(D[i]);
+    Safefree(D);
+  }
+
+  /* Estimate number of bits, from Pari, also see Stein 2006 */
+  nbits = mpz_logn(den) + (n+0.5) * log((double)n) - n*2.8378770664093454835606594728L + 1.712086L;
+  nbits /= log(2);
+  nbits += 32;
+  prec = (unsigned long)(nbits/3.32193 + 1);
+
+  mpf_init2(bn, nbits);
+  mpf_init2(tf, nbits);
+  _bern_real_zeta(bn, zn, prec);
+  mpf_set_z(tf, den);
+  mpf_mul(bn, bn, tf);
+
+  mpf_set_d(tf, (mpf_sgn(bn) < 0) ? -0.5 : 0.5);
+  mpf_add(bn, bn, tf);
+
+  mpz_set_f(num, bn);
+
+  mpf_clear(tf);
+  mpf_clear(bn);
+}
+
+
+void bernfrac(mpz_t num, mpz_t den, mpz_t zn)
+{
+  mpz_t t;
+  mpz_init(t);
+
+  if (mpz_cmp_ui(zn,46) < 0) {
+    _bernfrac_comb(num, den, zn, t);
+  } else {
+    _bernfrac_zeta(num, den, zn, t);
+  }
+
+  mpz_gcd(t, num, den);
+  mpz_divexact(num, num, t);
+  mpz_divexact(den, den, t);
+  mpz_clear(t);
+}
+
 char* bernreal(mpz_t zn, unsigned long prec) {
   char* out;
-  mpz_t num, den;
-
-  mpz_init(num); mpz_init(den);
-  bernfrac(num, den, zn);
-  out = frac_real(num, den, prec);
-  mpz_clear(den); mpz_clear(num);
 
+  if (mpz_cmp_ui(zn,40) < 0) {
+    mpz_t num, den, t;
+    mpz_init(num); mpz_init(den); mpz_init(t);
+    _bernfrac_comb(num, den, zn, t);
+    out = _frac_real(num, den, prec);
+    mpz_clear(t); mpz_clear(den); mpz_clear(num);
+  } else {
+    mpf_t z;
+    unsigned long bits = 32+(unsigned long)(prec*3.32193);
+    mpf_init2(z, bits);
+    _bern_real_zeta(z, zn, prec);
+    out = _str_real(z, prec);
+    mpf_clear(z);
+  }
   return out;
 }
 
+/***************************       Lambert W       ***************************/
+
+static void _lambertw(mpf_t w, mpf_t x, unsigned long prec)
+{
+  int i;
+  unsigned long bits = 96+mpf_get_prec(x);  /* More bits for intermediate */
+  mpf_t t, w1, zn, qn, en, tol;
+
+  if (mpf_cmp_d(x, -0.36787944117145) < 0)
+    croak("Invalid input to LambertW:  x must be >= -1/e");
+  if (mpf_sgn(x) == 0)
+    { mpf_set(w, x); return; }
+
+  /* Use Fritsch rather than Halley. */
+  mpf_init2(t,   bits);
+  mpf_init2(w1,  bits);
+  mpf_init2(zn,  bits);
+  mpf_init2(qn,  bits);
+  mpf_init2(en,  bits);
+  mpf_init2(tol, bits);
+
+  /* Initial estimate */
+  if (mpf_cmp_d(x, -0.06) < 0) {  /* Pade(3,2) */
+    mpf_set_d(t, 5.4365636569180904707205749);
+    mpf_mul(t, t, x);
+    mpf_add_ui(t, t, 2);
+    if (mpf_sgn(t) <= 0) { mpf_set_ui(t, 0); } else { mpf_sqrt(t, t); }
+    mpf_mul(zn, t, t);
+    mpf_mul(qn, zn, t);
+
+    mpf_set_d(w, -1);
+    mpf_set_d(w1, 1.0L/6.0L);      mpf_mul(w1, w1,  t);  mpf_add(w, w, w1);
+    mpf_set_d(w1, 257.0L/720.0L);  mpf_mul(w1, w1, zn);  mpf_add(w, w, w1);
+    mpf_set_d(w1, 13.0L/720.0L);   mpf_mul(w1, w1, zn);  mpf_add(w, w, w1);
+    mpf_set(en, w);  /* numerator */
+
+    mpf_set_d(w, 1);
+    mpf_set_d(w1, 5.0L/6.0L);      mpf_mul(w1, w1,  t);  mpf_add(w, w, w1);
+    mpf_set_d(w1, 103.0L/720.0L);  mpf_mul(w1, w1, zn);  mpf_add(w, w, w1);
+
+    mpf_div(w, en, w);
+  } else if (mpf_cmp_d(x, 1.363) < 0) {  /* Winitzki 2003 */
+    mpf_add_ui(t, x, 1);
+    mpf_log(w1, t);
+    mpf_add_ui(zn, w1, 1);
+    mpf_log(zn, zn);
+    mpf_add_ui(qn, w1, 2);
+    mpf_div(t, zn, qn);
+    mpf_ui_sub(t, 1, t);
+    mpf_mul(w, w1, t);
+  } else if (mpf_cmp_d(x, 3.7) < 0) {  /* Vargas 2013 modified */
+    mpf_log(w, x);
+    mpf_log(w1, w);
+    mpf_div(t, w1, w);
+    mpf_ui_sub(t, 1, t);
+    mpf_log(t, t);
+    mpf_div_ui(t, t, 2);
+    mpf_sub(w, w, w1);
+    mpf_sub(w, w, t);
+  } else {  /* Corless et al. 1993 */
+    mpf_t l1, l2, d1, d2, d3;
+    mpf_init2(l1,bits); mpf_init2(l2,bits);
+    mpf_init2(d1,bits); mpf_init2(d2,bits); mpf_init2(d3,bits);
+
+    mpf_log(l1, x);
+    mpf_log(l2, l1);
+    mpf_mul(d1, l1, l1);  mpf_mul_ui(d1, d1, 2);
+    mpf_mul(d2, l1, d1);  mpf_mul_ui(d2, d2, 3);
+    mpf_mul(d3, l1, d2);  mpf_mul_ui(d3, d3, 2);
+
+    mpf_sub(w, l1, l2);
+
+    mpf_div(t, l2, l1);
+    mpf_add(w, w, t);
+
+    mpf_sub_ui(t, l2, 2);
+    mpf_mul(t, t, l2);
+    mpf_div(t, t, d1);
+    mpf_add(w, w, t);
+
+    mpf_mul_ui(t, l2, 2);
+    mpf_sub_ui(t, t, 9);
+    mpf_mul(t, t, l2);
+    mpf_add_ui(t, t, 6);
+    mpf_mul(t, t, l2);
+    mpf_div(t, t, d2);
+    mpf_add(w, w, t);
+
+    mpf_mul_ui(t, l2, 3);
+    mpf_sub_ui(t, t, 22);
+    mpf_mul(t, t, l2);
+    mpf_add_ui(t, t, 36);
+    mpf_mul(t, t, l2);
+    mpf_sub_ui(t, t, 12);
+    mpf_mul(t, t, l2);
+    mpf_div(t, t, d3);
+    mpf_add(w, w, t);
+
+    mpf_clear(l1); mpf_clear(l2);
+    mpf_clear(d1); mpf_clear(d2); mpf_clear(d3);
+  }
+
+  /* Divide prec by 2 since t should be have 4x number of zeros each round */
+  mpf_set_ui(tol, 10);
+  mpf_pow_ui(tol, tol, (mpf_cmp_d(x, -.36) < 0) ? prec : prec/2);
+  mpf_ui_div(tol,1,tol);
+
+  for (i = 0; i < 500 && mpz_sgn(w) != 0; i++) {
+    mpf_add_ui(w1, w, 1);
+
+    mpf_div(t, x, w);
+    mpf_log(zn, t);
+    mpf_sub(zn, zn, w);
+
+    mpf_mul_ui(t, zn, 2);
+    mpf_div_ui(t, t, 3);
+    mpf_add(t, t, w1);
+    mpf_mul(t, t, w1);
+    mpf_mul_ui(qn, t, 2);
+
+    mpf_sub(en, qn, zn);
+    mpf_mul_ui(t, zn, 2);
+    mpf_sub(t, qn, t);
+    mpf_div(en, en, t);
+    mpf_div(t, zn, w1);
+    mpf_mul(en, en, t);
+
+    mpf_mul(t, w, en);
+    mpf_add(w, w, t);
+
+    mpf_abs(t, t);
+    if (mpf_cmp(t, tol) <= 0) break;
+    if (mpf_cmp_d(w,-1) <= 0) break;
+  }
+
+  if (mpf_cmp_d(w, -1) <= 0)
+    mpf_set_si(w, -1);
+
+  mpf_clear(en); mpf_clear(qn); mpf_clear(zn);
+  mpf_clear(w1); mpf_clear(t); mpf_clear(tol);
+}
+
+char* lambertwreal(mpf_t x, unsigned long prec) {
+  char* out;
+  mpf_t w;
+  unsigned long bits = 64+(unsigned long)(prec*3.32193);
+
+  mpf_init2(w, bits);
+  _lambertw(w, x, 10+prec);
+  out = _str_real(w, prec);
+  mpf_clear(w);
+  return out;
+}
+
+/*****************************************************************************/
+
 void stirling(mpz_t r, unsigned long n, unsigned long m, UV type)
 {
   mpz_t t, t2;
@@ -628,38 +1241,57 @@ void stirling(mpz_t r, unsigned long n, unsigned long m, UV type)
   } else {
     mpz_init(t);  mpz_init(t2);
     mpz_set_ui(r,0);
-    if (type == 3) {
-      mpz_bin_uiui(t, n-1, m-1);
-      mpz_fac_ui(t2, n);
+    if (type == 3) { /* Lah: binomial(n k) * binomial(n-1 k-1) * (n-k)!*/
+      mpz_bin_uiui(t, n, m);
+      mpz_bin_uiui(t2, n-1, m-1);
       mpz_mul(r, t, t2);
-      mpz_fac_ui(t2, m);
-      mpz_divexact(r, r, t2);
+      mpz_fac_ui(t2, n-m);
+      mpz_mul(r, r, t2);
     } else if (type == 2) {
-      for (j = 1; j <= m; j++) {
-        mpz_bin_uiui(t, m, j);
-        mpz_ui_pow_ui(t2, j, n);
-        mpz_mul(t, t, t2);
-        if ((m-j) & 1)  mpz_sub(r, r, t);
-        else            mpz_add(r, r, t);
+      mpz_t binom;
+      mpz_init_set_ui(binom, m);
+      mpz_ui_pow_ui(r, m, n);
+      /* Use symmetry to halve the number of loops */
+      for (j = 1; j <= ((m-1)>>1); j++) {
+        mpz_ui_pow_ui(t, j, n);
+        mpz_ui_pow_ui(t2, m-j, n);
+        if (m&1) mpz_sub(t, t2, t);
+        else     mpz_add(t, t2, t);
+        mpz_mul(t, t, binom);
+        if (j&1) mpz_sub(r, r, t);
+        else     mpz_add(r, r, t);
+        mpz_mul_ui(binom, binom, m-j);
+        mpz_divexact_ui(binom, binom, j+1);
       }
+      if (!(m&1)) {
+        mpz_ui_pow_ui(t, j, n);
+        mpz_mul(t, t, binom);
+        if (j&1) mpz_sub(r, r, t);
+        else     mpz_add(r, r, t);
+      }
+      mpz_clear(binom);
       mpz_fac_ui(t, m);
       mpz_divexact(r, r, t);
     } else {
+      mpz_bin_uiui(t,  n-1+1, n-m+1);
+      mpz_bin_uiui(t2, n-m+n, n-m-1);
+      mpz_mul(t2, t2, t);
       for (j = 1; j <= n-m; j++) {
-        mpz_bin_uiui(t,  n+j-1, n+j-m);
-        mpz_bin_uiui(t2, n+n-m, n-j-m);
-        mpz_mul(t, t, t2);
-        stirling(t2, n+j-m, j, 2);
+        stirling(t, n-m+j, j, 2);
         mpz_mul(t, t, t2);
         if (j & 1)      mpz_sub(r, r, t);
         else            mpz_add(r, r, t);
+        mpz_mul_ui(t2, t2, n+j);
+        mpz_divexact_ui(t2, t2, n-m+j+1);
+        mpz_mul_ui(t2, t2, n-m-j);
+        mpz_divexact_ui(t2, t2, n+j+1);
       }
     }
     mpz_clear(t2);  mpz_clear(t);
   }
 }
 
-/* Goetgheluck method.  Also thanks to Peter Luschny */
+/* Goetgheluck method.  Also thanks to Peter Luschny. */
 void binomial(mpz_t r, UV n, UV k)
 {
   UV fi, nk, sqrtn, piN, prime, i, j;
@@ -713,6 +1345,44 @@ void binomial(mpz_t r, UV n, UV k)
   Safefree(mprimes);
 }
 
+void partitions(mpz_t npart, UV n)
+{
+  mpz_t psum, *part;
+  UV *pent, i, j, k, d = (UV) sqrt(n+1);
+
+  if (n <= 3) {
+    mpz_set_ui(npart, (n == 0) ? 1 : n);
+    return;
+  }
+
+  New(0, pent, 2*d+2, UV);
+  pent[0] = 0;
+  pent[1] = 1;
+  for (i = 1; i <= d; i++) {
+    pent[2*i  ] = ( i   *(3*i+1)) / 2;
+    pent[2*i+1] = ((i+1)*(3*i+2)) / 2;
+  }
+  New(0, part, n+1, mpz_t);
+  mpz_init_set_ui(part[0], 1);
+  mpz_init(psum);
+  for (j = 1; j <= n; j++) {
+    mpz_set_ui(psum, 0);
+    for (k = 1; pent[k] <= j; k++) {
+      if ((k+1) & 2) mpz_add(psum, psum, part[ j - pent[k] ]);
+      else           mpz_sub(psum, psum, part[ j - pent[k] ]);
+    }
+    mpz_init_set(part[j], psum);
+  }
+
+  mpz_set(npart, part[n]);
+
+  mpz_clear(psum);
+  for (i = 0; i <= n; i++)
+    mpz_clear(part[i]);
+  Safefree(part);
+  Safefree(pent);
+}
+
 
 void _GMP_lcm_of_consecutive_integers(UV B, mpz_t m)
 {
@@ -759,7 +1429,7 @@ void _GMP_lcm_of_consecutive_integers(UV B, mpz_t m)
 /* a=0, return power.  a>1, return bool if an a-th power */
 UV is_power(mpz_t n, UV a)
 {
-  if (mpz_cmp_ui(n,3) <= 0)
+  if (mpz_cmp_ui(n,3) <= 0 && a == 0)
     return 0;
   else if (a == 1)
     return 1;
@@ -805,21 +1475,29 @@ static void word_tile(uint32_t* source, uint32_t from, uint32_t to) {
     from += words;
   }
 }
-static void sievep(uint32_t* comp, mpz_t start, UV p, UV len) {
-  UV pos = p - mpz_fdiv_ui(start,p);  /* First multiple of p after start  */
-  if (!(pos & 1)) pos += p;           /* Make sure it is odd.             */
-  for ( ; pos < len; pos += 2*p )
-    SETAVAL(comp, pos);
-}
-static void sievep_ui(uint32_t* comp, UV pos, UV p, UV len) {
+static void sievep_ui(uint32_t* comp, UV pos, UV p, UV len, int verbose) {
   if (!(pos & 1)) pos += p;
-  for ( ; pos < len; pos += 2*p )
-    SETAVAL(comp, pos);
+  if (verbose > 3) {
+    for ( ; pos < len; pos += 2*p ) {
+      if (!TSTAVAL(comp, pos)) {
+        printf("factor: %"UVuf" at %"UVuf"\n", p, pos);
+        SETAVAL(comp, pos);
+      }
+    }
+  } else {
+    for ( ; pos < len; pos += 2*p )
+      SETAVAL(comp, pos);
+  }
 }
+/* Find first multiple of p after start */
+#define sievep(comp, start, p, len, verbose) \
+  sievep_ui(comp, (p) - mpz_fdiv_ui((start),(p)), p, len, verbose)
+
 uint32_t* partial_sieve(mpz_t start, UV length, UV maxprime)
 {
   uint32_t* comp;
   UV p, wlen, pwlen;
+  int _verbose = get_verbose_level();
   PRIME_ITERATOR(iter);
 
   /* mpz_init(t);
@@ -839,7 +1517,7 @@ uint32_t* partial_sieve(mpz_t start, UV length, UV maxprime)
   pwlen = (wlen < 3) ? wlen : 3;
   memset(comp, 0x00, pwlen*sizeof(uint32_t));
   while (p <= maxprime) {
-    sievep(comp, start, p, pwlen*64);
+    sievep(comp, start, p, pwlen*64, _verbose);
     p = prime_iterator_next(&iter);
     if (pwlen*p >= wlen) break;
     word_tile(comp, pwlen, pwlen*p);
@@ -865,12 +1543,12 @@ uint32_t* partial_sieve(mpz_t start, UV length, UV maxprime)
           p1 = prime_iterator_next(&iter), p2 = prime_iterator_next(&iter) ) {
       UV p1p2 = p1 * p2;
       UV ddiv = mpz_fdiv_ui(start, p1p2);
-      sievep_ui(comp, p1 - (ddiv % p1), p1, length);
-      sievep_ui(comp, p2 - (ddiv % p2), p2, length);
+      sievep_ui(comp, p1 - (ddiv % p1), p1, length, _verbose);
+      sievep_ui(comp, p2 - (ddiv % p2), p2, length, _verbose);
     }
-    if (p1 <= maxprime) sievep(comp, start, p1, length);
+    if (p1 <= maxprime) sievep(comp, start, p1, length, _verbose);
     for (p = p2; p <= ulim; p = prime_iterator_next(&iter))
-      sievep(comp, start, p, length);
+      sievep(comp, start, p, length, _verbose);
     if (p < maxprime) {
       /* UV is 64-bit, GMP's ui functions are 32-bit.  Sigh. */
       UV lastp, pos;
@@ -889,7 +1567,7 @@ uint32_t* partial_sieve(mpz_t start, UV length, UV maxprime)
           p1 += ((UV)mpz_get_ui(rem)) << 31;
           pos = p - p1;
         }
-        sievep_ui(comp, pos, p, length);
+        sievep_ui(comp, pos, p, length, _verbose);
       }
       mpz_clear(mp);
       mpz_clear(rem);
@@ -1085,7 +1763,7 @@ typedef struct {
 
 UV* sieve_primes(mpz_t inlow, mpz_t high, UV k, UV *rn) {
   mpz_t t, low;
-  int test_primality = 0, k_primality = 0;
+  int test_primality = 0, k_primality = 0, force_full = 0, width2, hbits;
   uint32_t* comp;
   vlist retlist;
 
@@ -1093,16 +1771,25 @@ UV* sieve_primes(mpz_t inlow, mpz_t high, UV k, UV *rn) {
   if (mpz_cmp(inlow, high) > 0) { *rn = 0; return 0; }
 
   mpz_init(t);
+  mpz_sub(t, high, inlow);
+  width2 = mpz_sizeinbase(t,2);
+  hbits = mpz_sizeinbase(high,2);
+
   mpz_sqrt(t, high);           /* No need for k to be > sqrt(high) */
   /* If auto-setting k or k >= sqrt(n), pick a good depth and test primality */
   if (k == 0 || mpz_cmp_ui(t, k) <= 0) {
-    /* TODO: We need to take into account the sieve width */
-    UV hbits = mpz_sizeinbase(high,2);
     test_primality = 1;
-    k = (hbits < 100) ? 50000000 : hbits*500000;
+    k = (hbits < 100) ? 50000000UL : ((UV)hbits)*UVCONST(500000);
   }
+  /* For smallist n and large ranges, force full sieve */
+  if (test_primality && hbits >= 40 && hbits <  56 && (width2 * 2.8) >= hbits)
+    force_full = 1;
+  if (test_primality && hbits >= 56 && hbits <  64 && (width2 * 2.6) >= hbits)
+    force_full = 1;
+  if (test_primality && hbits >= 64 && hbits <= 82 && (width2 * 2.5) >= hbits && sizeof(unsigned long) > 4)
+    force_full = 1;
   /* If k >= sqrtn, sieving is enough.  Use k=sqrtn, turn off post-sieve test */
-  if (mpz_cmp_ui(t, k) <= 0) {
+  if (force_full || mpz_cmp_ui(t, k) <= 0) {
     k = mpz_get_ui(t);
     k_primality = 1;           /* Our sieve is complete */
     test_primality = 0;        /* Don't run BPSW */
