@@ -5,7 +5,7 @@ use Carp qw/croak confess carp/;
 
 BEGIN {
   $Math::Prime::Util::GMP::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::GMP::VERSION = '0.40';
+  $Math::Prime::Util::GMP::VERSION = '0.42';
 }
 
 # parent is cleaner, and in the Perl 5.10.1 / 5.12.0 core, but not earlier.
@@ -23,6 +23,7 @@ our @EXPORT_OK = qw(
                      is_bls75_prime
                      is_ecpp_prime
                      is_pseudoprime
+                     is_euler_pseudoprime
                      is_euler_plumb_pseudoprime
                      is_strong_pseudoprime
                      is_lucas_pseudoprime
@@ -57,6 +58,7 @@ our @EXPORT_OK = qw(
                      ecm_factor
                      qs_factor
                      factor
+                     divisors
                      sigma
                      chinese
                      moebius
@@ -66,6 +68,7 @@ our @EXPORT_OK = qw(
                      factorial
                      consecutive_integer_lcm
                      partitions bernfrac bernreal harmfrac harmreal stirling
+                     zeta riemannr lambertw
                      gcd lcm kronecker valuation binomial gcdext
                      invmod sqrtmod addmod mulmod divmod powmod
                      vecsum vecprod
@@ -75,31 +78,53 @@ our @EXPORT_OK = qw(
                      jordan_totient
                      carmichael_lambda
                      sqrtint rootint
-                     is_power is_prime_power
+                     is_power is_prime_power is_semiprime
                      is_primitive_root
                      znorder
                      znprimroot
                      ramanujan_tau
                      Pi
+                     todigits
+                     random_prime random_nbit_prime random_ndigit_prime
+                     seed_csprng is_csprng_well_seeded
                    );
                    # Should add:
                    # nth_prime
 our %EXPORT_TAGS = (all => [ @EXPORT_OK ]);
 
+sub _init_random {
+  use Fcntl;
+  my($file, $nbytes, $s, $buffer, $nread) = ("/dev/urandom", 256, '', '', 0);
+  return if $^O eq 'MSWin32';
+  return unless -r $file;
+  sysopen(my $fh, $file, O_RDONLY);
+  binmode $fh;
+  while ($nread < $nbytes) {
+    my $thisread = sysread $fh, $buffer, $nbytes-$nread;
+    last unless defined $thisread && $thisread > 0;
+    $s .= $buffer;
+    $nread += length($buffer);
+  }
+  return unless $nbytes == length($s);
+  seed_csprng($nbytes, $s);
+}
+
 BEGIN {
   eval {
     require XSLoader;
     XSLoader::load(__PACKAGE__, $Math::Prime::Util::GMP::VERSION);
     _GMP_init();
+    _init_random();
     1;
   } or do {
     die $@;
-  }
+  };
 }
 END {
   _GMP_destroy();
 }
 
+
 sub _validate_positive_integer {
   my($n, $min, $max) = @_;
   croak "Parameter must be defined" if !defined $n;
@@ -116,17 +141,6 @@ sub _validate_positive_integer {
 }
 
 
-sub is_strong_pseudoprime {
-  my($n, @bases) = @_;
-  _validate_positive_integer($n);
-  croak "No bases given to is_strong_pseudoprime" unless @bases;
-  foreach my $base (@bases) {
-    _validate_positive_integer($base);
-    return 0 unless _GMP_miller_rabin("$n", "$base");
-  }
-  1;
-}
-
 sub is_provable_prime {
   my ($n) = @_;
   return 0 if $n < 2;
@@ -147,13 +161,6 @@ sub is_provable_prime_with_cert {
   return ($result, $text);
 }
 
-sub factor {
-  my ($n) = @_;
-  my @factors = ($n < 4) ? ($n)
-                         : sort {$a<=>$b} _GMP_factor($n);
-  return @factors;
-}
-
 sub primes {
   my($low,$high) = (scalar(@_) == 1) ? (2,$_[0]) : ($_[0], $_[1]);
 
@@ -174,7 +181,7 @@ __END__
 
 =encoding utf8
 
-=for stopwords Möbius Deléglise Bézout gcdext vecsum vecprod moebius totient liouville znorder znprimroot bernfrac bernreal harmfrac harmreal stirling lucasu lucasv OpenPFGW gmpy2 nonresidue chinese tuplets sqrtmod addmod mulmod powmod divmod superset sqrtint rootint
+=for stopwords Möbius Deléglise Bézout gcdext vecsum vecprod moebius totient liouville znorder znprimroot bernfrac bernreal harmfrac harmreal stirling zeta riemannr lambertw lucasu lucasv OpenPFGW gmpy2 nonresidue chinese tuplets sqrtmod addmod mulmod powmod divmod superset sqrtint rootint todigits
 
 =head1 NAME
 
@@ -183,7 +190,7 @@ Math::Prime::Util::GMP - Utilities related to prime numbers and factoring, using
 
 =head1 VERSION
 
-Version 0.40
+Version 0.42
 
 
 =head1 SYNOPSIS
@@ -309,8 +316,7 @@ This call walks the line between the performance of L</is_prob_prime>
 and the certainty of L</is_provable_prime>.  Those calls may be more
 appropriate in some cases.  What this function does is give most of
 the performance of the former, while adding more certainty.  For finer
-tuning of this tradeoff, especially if performance is critical for
-65- to 200-bit inputs, you may instead use L</is_prob_prime> followed
+tuning of this tradeoff, you may instead use L</is_prob_prime> followed
 by additional probable prime tests such as L</miller_rabin_random>
 and/or L</is_frobenius_underwood_pseudoprime>.
 
@@ -325,8 +331,8 @@ still "probably prime" but under 200 bits, a quick
 BLS75 C<n-1> primality proof is attempted.  This is tuned to give up
 if the result cannot be quickly determined, and results in success
 rates of ~80% at 80 bits, ~30% at 128 bits, and ~13% at 160 bits.
-Lastly, for results still "probably prime", a small number of
-Miller-Rabin tests with random bases are performed.
+Lastly, for results still "probably prime", an additional random-base
+Miller-Rabin test is performed.
 
 The result is that many numbers will return 2 (definitely prime),
 and the numbers that return 1 (probably prime) have gone through
@@ -398,20 +404,26 @@ verification.  Proof types used include:
 
 =head2 is_pseudoprime
 
-Takes a positive number C<n> and a base C<a> as input, and returns 1 if
-C<n> is a probable prime to base C<a>.  This is the simple Fermat primality
-test.  Removing primes, given base 2 this produces the sequence
-L<OEIS A001567|http://oeis.org/A001567>.
+Takes a positive number C<n> and one or more non-zero positive bases as input.
+Returns C<1> if the input is a probable prime to each base, C<0> if not.
+This is the simple Fermat primality test.
+Removing primes, given base 2 this produces the sequence L<OEIS A001567|http://oeis.org/A001567>.
+
+=head2 is_euler_pseudoprime
+
+Takes a positive number C<n> and one or more non-zero positive bases as input.
+Returns C<1> if the input is an Euler probable prime to each base, C<0> if not.
+This is the Euler test, sometimes called the Euler-Jacobi test.
+Removing primes, given base 2 this produces the sequence L<OEIS A047713|http://oeis.org/A047713>.
 
 =head2 is_strong_pseudoprime
 
   my $maybe_prime = is_strong_pseudoprime($n, 2);
   my $probably_prime = is_strong_pseudoprime($n, 2, 3, 5, 7, 11, 13, 17);
 
-Takes a positive number as input and one or more bases.  Returns 1 if
-the input is a prime or a strong pseudoprime to all of the bases, and
-0 if not.  The base must be a positive integer.  This is often called
-the Miller-Rabin test.
+Takes a positive number C<n> and one or more non-zero positive bases as input.
+Returns C<1> if the input is a strong probable prime to each base, C<0> if not.
+This is often called the Miller-Rabin test.
 
 If 0 is returned, then the number really is a composite.  If 1 is
 returned, then it is either a prime or a strong pseudoprime to all
@@ -894,6 +906,34 @@ Note that with a non-zero second argument, the values returned have not
 undergone a full BPSW test; just sieving and a SPSP-2 test.
 
 
+=head2 random_nbit_prime
+
+  say "random 512-bit prime: ", random_nbit_prime(512);
+
+Returns a randomly selected prime of exactly C<n> bits.
+C<undef> is returned if C<n> is less than C<2>.
+The returned prime has passed the C<is_prob_prime> (extra strong BPSW) test.
+
+=head2 random_ndigit_prime
+
+  say "random 200-digit prime: ", random_ndigit_prime(200);
+
+Returns a randomly selected prime of exactly C<n> digits.
+C<undef> is returned if C<n> is less than C<1>.
+The returned prime has passed the C<is_prob_prime> (extra strong BPSW) test.
+
+=head2 random_prime
+
+  say random_prime(1000, 2000);  # prime between 1000 and 2000 inclusive
+
+Returns a random prime in the interval C<[a,b]> or C<undef> if no prime is
+in the range.
+The returned prime has passed the C<is_prob_prime> (extra strong BPSW) test.
+
+The random prime functions use the internal CSPRNG for randomness.  This
+is currently ISAAC-32.
+
+
 =head2 lucasu
 
   say "Fibonacci($_) = ", lucasu(1,-1,$_) for 0..100;
@@ -1045,7 +1085,8 @@ functions.
 
 Returns the Bernoulli number C<B_n> for an integer argument C<n>, as a
 string floating point.  An optional second argument indicates the number
-of digits to be preserved past the decimal place, with a default of 40.
+of significant digits to be used, with the result rounded.  The default
+is 40 digits.
 
 =head2 harmfrac
 
@@ -1078,6 +1119,32 @@ of the second kind are the number of ways to partition a set of C<n>
 elements into C<k> non-empty subsets.  The Lah numbers are the number of
 ways to split a set of C<n> elements into C<k> non-empty lists.
 
+=head2 zeta
+
+Given a positive integer or float C<n>, returns the real Riemann Zeta
+value as a string floating point.  An optional second argument indicates
+the number of digits past the decimal point (default 40).
+
+The implementation is algorithm 2 of Borwein (1991).  Performance with
+integer inputs is good, but floating point arguments with high precision
+will be much slower than methods using MPFR.  L<Math::Prime::Util> will
+try to use L<Math::MPFR> if possible.
+
+=head2 riemannr
+
+Given a positive integer or float C<n>, returns the real Riemann R function
+as a string floating point.  An optional second argument indicates the number
+of significant digits (default 40) with the result rounded.
+
+The implementation is the standard Gram series.
+
+=head2 lambertw
+
+Given a float C<x>, returns the principal branch of the Lambert W function.
+This solves for C<W> in the equation C<x = W*exp(W)>.  The input must not be
+less than C<-1/e>.  This corresponds to Pari's C<lambertw>
+function and Mathematica's C<ProductLog> / C<LambertW> function.
+
 
 =head2 znorder
 
@@ -1110,6 +1177,14 @@ Given two non-negative numbers C<a> and C<n>, returns C<1> if C<a> is a
 primitive root modulo C<n>, and C<0> if not.  If C<a> is a primitive root,
 then C<euler_phi(n)> is the smallest C<e> for which C<a^e = 1 mod n>.
 
+=head2 is_semiprime
+
+Given a positive integer C<n>, returns 1 if C<n> is a semiprime, 0 otherwise.
+A semiprime is the product of exactly two primes.
+
+The boolean result is the same as C<scalar(factor(n)) == 2>, but this
+function performs shortcuts that can greatly speed up the operation.
+
 
 =head2 sigma
 
@@ -1374,6 +1449,23 @@ L<gmp-ecm|http://ecm.gforge.inria.fr/>, and
 L<GGNFS|http://sourceforge.net/projects/ggnfs/>.
 
 
+=head2 divisors
+
+  my @divisors = divisors(30);   # returns (1, 2, 3, 5, 6, 10, 15, 30)
+
+Produces all the divisors of a positive number input, including 1 and
+the input number.  The divisors are a power set of multiplications of
+the prime factors, returned as a sorted list with no duplications.
+The result is identical to that of Pari's C<divisors> and
+Mathematica's C<Divisors[n]> functions.
+
+In scalar context this returns the sigma0 function (OEIS A000005), and
+is the same result as evaluating the returned array in scalar context
+(but much more efficient).
+The result then corresponds to Pari's C<numdiv> and Mathematica's
+C<DivisorSigma[0,n]> functions.
+
+
 =head2 trial_factor
 
   my @factors = trial_factor($n);
@@ -1579,6 +1671,42 @@ However, it is substantially faster than the other methods on large inputs
 having large factors, and is the method of choice for 35+ digit semiprimes.
 
 
+=head2 todigits
+
+Given an integer C<n>, return an array of digits of C<|n|>.  An optional
+second integer argument specifies a base (default 10).  For example,
+given a base of 2, this returns an array of binary digits of C<n>.
+An optional third argument specifies a length for the returned array.
+The result will be either have upper digits truncated or have leading
+zeros added.
+
+C<todigits(0)> returns an empty array.
+The base must be at least 2, and is limited to an int.
+Length must be at least zero and is limited to an int.
+
+
+=head2 seed_csprng
+
+Takes a non-negative integer C<nbytes> and a string C<data> as input.
+These are used to seed the internal CSPRNG used for random functions,
+including the random prime functions.  Ideally this is 16-256 bytes of
+good entropy.
+
+Currently the CSPRNG is ISAAC-32, and the maximum number of seed bytes
+used is 1024.  It is possible the CSPRNG used will change in the future
+(thoughts include Salsa20, SHA2-256 counter, or SHA3-256 counter).
+
+
+=head2 is_csprng_well_seeded
+
+Returns true if the CSPRNG has been seeded with 16 or more bytes (128 bits).
+There is no measurement of how "good" the input was.
+
+On startup the module will attempt to seed the CSPRNG from
+C</dev/urandom>, so this function will return true if that was
+successful, but false otherwise.
+
+
 =head1 SEE ALSO
 
 =over 4
@@ -1696,7 +1824,7 @@ ECM implementation, as well as the papers by Brent and Montgomery.
 
 =head1 COPYRIGHT
 
-Copyright 2011-2016 by Dana Jacobsen E<lt>dana@acm.orgE<gt>
+Copyright 2011-2017 by Dana Jacobsen E<lt>dana@acm.orgE<gt>
 
 This program is free software; you can redistribute it and/or modify it under the same terms as Perl itself.
 
