@@ -12,6 +12,8 @@
 /* includes mpz_mulmod(r, a, b, n, temp) */
 #include "utility.h"
 #include "factor.h"
+#include "primality.h"
+#include "isaac.h"
 
 static int _verbose = 0;
 int get_verbose_level(void) { return _verbose; }
@@ -27,14 +29,61 @@ void init_randstate(unsigned long seed) {
   gmp_randinit_default(_randstate);
 #endif
   gmp_randseed_ui(_randstate, seed);
+  isaac_init( sizeof(unsigned long), (const unsigned char*)(&seed) );
 }
 void clear_randstate(void) {  gmp_randclear(_randstate);  }
 
+void mpz_isaac_urandomb(mpz_t rop, int nbits)
+{
+  unsigned char* d;
+  int nbytes = (nbits+7)/8;
+  New(0, d, nbytes, unsigned char);
+  isaac_rand_bytes(nbytes, d);
+  mpz_import(rop, nbytes, 1, sizeof(unsigned char), 0, 0, d);
+  Safefree(d);
+  if (nbits != nbytes*8)
+    mpz_tdiv_r_2exp(rop, rop, nbits);
+}
+
+void mpz_isaac_urandomm(mpz_t rop, mpz_t n)
+{
+  int count = 80;
+  unsigned long nbits = mpz_sizeinbase(n,2);
+
+  if (mpz_sgn(n) <= 0) {
+    mpz_set_ui(rop,0);
+    return;
+  } else if (nbits <= 32) {
+    mpz_set_ui(rop, isaac_rand(mpz_get_ui(n)));
+  } else if (nbits < 3000) {
+    /* Just like GMP, try until we're in range or we're tired. */
+    while (count-- > 0) {
+      mpz_isaac_urandomb(rop, nbits);
+      if (mpz_cmp(rop,n) < 0)
+        return;
+    }
+    mpz_mod(rop, rop, n);
+  } else {
+    /* Reduce tries needed by selecting from a range that is a multiple of n
+     * (so no bias) and uses the max space inside the power-of-2 range.
+     * Downside is that we do an alloc and two mods.  For large values
+     * it can be much faster however. */
+    mpz_t rmax;
+    mpz_init(rmax);
+    mpz_setbit(rmax, nbits+8);
+    mpz_sub_ui(rmax,rmax,1);
+    mpz_tdiv_q(rmax, rmax, n);
+    mpz_mul(rmax, rmax, n);
+    do {
+      mpz_isaac_urandomb(rop, nbits+8);
+    } while (mpz_cmp(rop, rmax) >= 0 && count-- > 0);
+    mpz_mod(rop, rop, n);
+  }
+}
 
 int is_primitive_root(mpz_t a, mpz_t n, int nprime)
 {
-  const unsigned char pr[10] = {2,3,5,7,11,13,17,19,23,29};
-  mpz_t s, t, *factors;
+  mpz_t s, sreduced, t, *factors;
   int ret, i, nfactors, *exponents;
 
   if (mpz_sgn(n) == 0)       return 0;
@@ -51,35 +100,66 @@ int is_primitive_root(mpz_t a, mpz_t n, int nprime)
   } else {
     totient(s, n);
   }
+  mpz_init_set(sreduced, s);
   mpz_init(t);
   ret = 1;
-  /* Try a few small primes first before complete factoring */
-  for (i = 0; i < 10; i++) {
-    unsigned long p = pr[i];
-    if (mpz_cmp_ui(s, p) < 0) break;
-    if (mpz_divisible_ui_p(s, p) ) {
-      mpz_divexact_ui(t, s, p);
-      mpz_powm(t, a, t, n);
-      if (mpz_cmp_ui(t, 1) == 0) {
-        ret = 0;
-        break;
-      }
+
+#define IPR_TEST_UI(s, p, a, n, t, ret) \
+  mpz_divexact_ui(t, s, p); \
+  mpz_powm(t, a, t, n); \
+  if (mpz_cmp_ui(t, 1) == 0) { ret = 0; }
+
+#define IPR_TEST(s, p, a, n, t, ret) \
+  mpz_divexact(t, s, p); \
+  mpz_powm(t, a, t, n); \
+  if (mpz_cmp_ui(t, 1) == 0) { ret = 0; }
+
+  { /* Pull out small factors and test */
+    UV p, fromp = 0;
+    while (ret == 1 && (p = _GMP_trial_factor(sreduced, fromp, 60))) {
+      if (mpz_cmp_ui(sreduced,p) == 0) break;
+      IPR_TEST_UI(s, p, a, n, t, ret);
+      mpz_set_ui(t, p);
+      (void) mpz_remove(sreduced, sreduced, t);
+      fromp = p+1;
     }
   }
-  if (ret == 1) {
-    nfactors = factor(s, &factors, &exponents);
-    for (i = 0; i < nfactors; i++) {
-      if (mpz_cmp_ui(factors[i],30) < 0) continue;
-      mpz_divexact(t, s, factors[i]);
-      mpz_powm(t, a, t, n);
-      if (mpz_cmp_ui(t,1) == 0) {
-        ret = 0;
-        break;
-      }
-    }
-    clear_factors(nfactors, &factors, &exponents);
+  if (ret == 0 || mpz_cmp_ui(sreduced,1) == 0)
+    goto DONE_IPR;
+  if (_GMP_BPSW(sreduced)) {
+    IPR_TEST(s, sreduced, a, n, t, ret);
+    goto DONE_IPR;
   }
-  mpz_clear(s);  mpz_clear(t);
+
+  /* Pull out more factors, noting they can be composites. */
+  while (_GMP_pminus1_factor(sreduced, t, 100000, 100000)) {
+    mpz_divexact(sreduced, sreduced, t);
+    if (_GMP_BPSW(t)) {
+      IPR_TEST(s, t, a, n, t, ret);
+    } else {
+      nfactors = factor(t, &factors, &exponents);
+      for (i = 0; ret == 1 && i < nfactors; i++) {
+        IPR_TEST(s, factors[i], a, n, t, ret);
+      }
+      clear_factors(nfactors, &factors, &exponents);
+    }
+    if (ret == 0 || mpz_cmp_ui(sreduced,1) == 0)
+      goto DONE_IPR;
+    if (_GMP_BPSW(sreduced)) {
+      IPR_TEST(s, sreduced, a, n, t, ret);
+      goto DONE_IPR;
+    }
+  }
+
+  /* We have a composite and so far it could be a primitive root.  Factor. */
+  nfactors = factor(sreduced, &factors, &exponents);
+  for (i = 0; ret == 1 && i < nfactors; i++) {
+    IPR_TEST(s, factors[i], a, n, t, ret);
+  }
+  clear_factors(nfactors, &factors, &exponents);
+
+DONE_IPR:
+  mpz_clear(s);  mpz_clear(sreduced);  mpz_clear(t);
   return ret;
 }
 
@@ -451,6 +531,11 @@ int chinese(mpz_t ret, mpz_t lcm, mpz_t *a, mpz_t *m, int items)
   }
 #endif
 
+  /* Avoid dividing by zero, which GMP doesn't enjoy. */
+  for (i = 0; i < items; i++)
+    if (mpz_sgn(m[i]) == 0)
+      return 0;
+
   mpz_init(temp1); mpz_init(temp2);
   mpz_init(sum); mpz_init(gcd);
   mpz_init(s); mpz_init(t);
@@ -538,6 +623,167 @@ void mpz_product(mpz_t* A, UV a, UV b) {
   }
 }
 
+/******************************************************************************/
+/*
+ * Floating point routines.
+ * These are very crude.  Use MPFR if at all possible.
+ *
+ */
+
+void mpf_log(mpf_t logn, mpf_t n)
+{
+  mpf_t N, a, b, t;
+  unsigned long k, bits = mpf_get_prec(n);
+
+  mpf_init2(N, bits);
+  mpf_init2(a, 64 + bits);
+  mpf_init2(b, 64 + bits);
+  mpf_init2(t, 64 + bits);
+
+  mpf_set(N, n);
+  mpf_set_ui(logn, 0);
+  mpf_set_ui(t, 1);
+  mpf_mul_2exp(t, t, 32);
+  for (k = 0; mpf_cmp(N, t) > 0; k++) {
+    mpf_div_2exp(N, N, 32);
+  }
+  if (k > 0) {
+    mpf_log(b, t);
+    mpf_mul_ui(logn, b, k);
+  }
+
+  mpf_set_ui(a, 1);
+  mpf_set(b, N);
+  /* TODO: this is AGM(1,N), make this generic */
+  mpf_mul(t, a, b);
+  mpf_add(a, a, b);
+  mpf_div_2exp(a, a, 1);
+  mpf_sqrt(b, t);
+  while (1) {
+    mpf_sub(t, a, b);
+    mpf_abs(t, t);
+    mpf_mul_2exp(t, t, bits);
+    if (mpf_cmp_d(t, .5) < 0)
+      break;
+    mpf_add(t, a, b);
+    mpf_div_2exp(a, t, 1);
+    mpf_mul(t, b, a);
+    mpf_sqrt(b, t);
+  }
+  /* End of generic AGM */
+  mpf_sub_ui(t, N, 1);
+  mpf_div(t, t, a);
+  mpf_add(logn, logn, t);
+
+  mpf_clear(t); mpf_clear(b); mpf_clear(a); mpf_clear(N);
+}
+
+void _mpf_lift_exp(mpf_t xj, mpf_t y, mpf_t t, mpf_t t2)
+{
+  unsigned long k;
+
+  mpf_log(t, xj);
+
+  mpf_sub(t2, y, t);      /* t2 = y-ln(xj) */
+  mpf_mul(t, xj, t2);     /* t = xj(y-ln(x)) */
+  mpf_add(xj, xj, t);
+
+  /* third and higher orders */
+  for (k = 3; k <= 8; k++) {
+    mpf_mul(t, t, t2);
+    mpf_div_ui(t, t, k-1);
+    mpf_add(xj, xj, t);
+  }
+}
+
+void mpf_exp(mpf_t expn, mpf_t x)
+{
+  mpf_t N, D, X, s, t;
+  unsigned long k, kinit, bits = mpf_get_prec(x);
+
+  mpf_init2(t, 64 + bits);
+
+  /* Doubling rule, to make -.25 < x < .25.  Speeds convergence. */
+  mpf_abs(t, x);
+  for (k = 0; k < 30 && mpf_cmp_d(t, .25) > 0; k++)
+    mpf_div_2exp(t, t, 1);
+  if (k > 0) {
+    if (mpf_sgn(x) < 0)
+      mpf_neg(t, t);
+
+    mpf_exp(expn, t);
+
+    mpf_pow_ui(expn, expn, 1 << k);
+    mpf_clear(t);
+    return;
+  }
+
+  mpf_init2(N, 64 + bits);
+  mpf_init2(D, 64 + bits);
+  mpf_init2(s, 64 + bits);
+  mpf_init2(X, 64 + bits);
+
+  /* 1. Compute s =~ sinh(x). */
+  mpf_set(s,x);
+  mpf_set(N,x);
+  mpf_mul(X,x,x);
+  mpf_set_ui(D,1);
+  kinit = bits/10;
+  for (k = 1; k < kinit; k++) {
+    mpf_mul(N, N, X);
+    mpf_mul_ui(D, D, 2*k);
+    mpf_mul_ui(D, D, 2*k+1);
+    mpf_div(t, N, D);
+    mpf_add(s, s, t);
+
+    mpf_abs(t, t);
+    mpf_mul_2exp(t, t, bits);
+    if (mpf_cmp_d(t, .5) < 0)
+      break;
+  }
+  mpf_clear(X);
+
+  /* 2. Compute s =~ e(x) from sinh(x). */
+  mpf_mul(t, s, s);
+  mpf_add_ui(t, t, 1);
+  mpf_sqrt(t, t);
+  mpf_add(s, s, t);
+
+  /* 3. If needed, use high-order Newton for more bits */
+  if (k >= kinit) {
+    for (k = 1; k < 100; k++) {
+      mpf_set(t, s);
+      _mpf_lift_exp(s, x, N, D);
+      mpf_sub(t, s, t);
+      mpf_abs(t,t);
+      mpf_mul_2exp(t, t, bits);
+      if (mpf_cmp_d(t, .5) < 0)
+        break;
+    }
+  }
+
+  mpf_set(expn, s);
+  mpf_clear(s); mpf_clear(D); mpf_clear(N); mpf_clear(t);
+}
+
+void mpf_pow(mpf_t powx, mpf_t b, mpf_t x)
+{
+  mpf_t t;
+
+  if (mpf_integer_p(x) && mpf_fits_ulong_p(x)) {
+    mpf_pow_ui(powx, b, mpf_get_ui(x));
+    return;
+  }
+
+  mpf_init2(t, mpf_get_prec(powx));
+  mpf_log(t, b);
+  mpf_mul(t, t, x);
+  mpf_exp(powx, t);
+  mpf_clear(t);
+}
+
+/******************************************************************************/
+
 
 #if 0
 /* Simple polynomial multiplication */
@@ -1066,8 +1312,7 @@ void polyz_root_deg2(mpz_t root1, mpz_t root2, mpz_t* pn, mpz_t NMOD)
  * Step 3/4 of Cohen Algorithm 1.6.1.
  * Uses some hints from Pate Williams (1997-1998) for the poly math */
 static void polyz_roots(mpz_t* roots, long *nroots,
-                        long maxroots, mpz_t* pg, long dg, mpz_t NMOD,
-                        gmp_randstate_t* p_randstate)
+                        long maxroots, mpz_t* pg, long dg, mpz_t NMOD)
 {
   long i, ntries, maxtries, maxd, dxa, dt, dh, dq, dup;
   mpz_t t, power;
@@ -1134,7 +1379,7 @@ static void polyz_roots(mpz_t* roots, long *nroots,
   while (ntries++ < maxtries) {
     /* pxa = X+a for randomly selected a */
     if (ntries <= 2)  mpz_set_ui(pxa[0], ntries);  /* Simple small values */
-    else              mpz_urandomm(pxa[0], *p_randstate, t);
+    else              mpz_isaac_urandomm(pxa[0], t);
 
     /* Raise pxa to (NMOD-1)/2, all modulo NMOD and g(x) */
     polyz_pow_polymod(pt, pxa, pg, &dt, dxa, dg, power, NMOD);
@@ -1150,17 +1395,17 @@ static void polyz_roots(mpz_t* roots, long *nroots,
   if (dh >= 1 && dh < dg) {
     /* Pick the smaller of the two splits to process first */
     if (dh <= 2 || dh <= (dg-dh)) {
-      polyz_roots(roots, nroots, maxroots, ph, dh, NMOD, p_randstate);
+      polyz_roots(roots, nroots, maxroots, ph, dh, NMOD);
       if (*nroots < maxroots) {
         /* q = g/h, and recurse */
         polyz_div(pq, pt,  pg, ph,  &dq, &dt, dg, dh, NMOD);
-        polyz_roots(roots, nroots, maxroots, pq, dq, NMOD, p_randstate);
+        polyz_roots(roots, nroots, maxroots, pq, dq, NMOD);
       }
     } else {
       polyz_div(pq, pt,  pg, ph,  &dq, &dt, dg, dh, NMOD);
-      polyz_roots(roots, nroots, maxroots, pq, dq, NMOD, p_randstate);
+      polyz_roots(roots, nroots, maxroots, pq, dq, NMOD);
       if (*nroots < maxroots) {
-        polyz_roots(roots, nroots, maxroots, ph, dh, NMOD, p_randstate);
+        polyz_roots(roots, nroots, maxroots, ph, dh, NMOD);
       }
     }
   }
@@ -1182,8 +1427,7 @@ static void polyz_roots(mpz_t* roots, long *nroots,
 
 /* Algorithm 1.6.1 from Cohen, minus step 1. */
 void polyz_roots_modp(mpz_t** roots, long *nroots, long maxroots,
-                      mpz_t *pP, long dP, mpz_t NMOD,
-                      gmp_randstate_t* p_randstate)
+                      mpz_t *pP, long dP, mpz_t NMOD)
 {
   long i;
 
@@ -1218,7 +1462,7 @@ void polyz_roots_modp(mpz_t** roots, long *nroots, long maxroots,
   if (maxroots > dP || maxroots == 0)
     maxroots = dP;
 
-  polyz_roots(*roots, nroots, maxroots, pP, dP, NMOD, p_randstate);
+  polyz_roots(*roots, nroots, maxroots, pP, dP, NMOD);
   /* This could be just really bad luck.  Let the caller handle it. */
   /* if (*nroots == 0) croak("failed to find roots\n"); */
 
